# coding: utf-8

from .core import Archetype, Page, RenderedString, settings
import re
import os
import io
import pytz
import datetime
import markdown
import dateutil.parser
from urllib.parse import urlparse, urlunparse
from .utils import parse_front_matter
import logging

log = logging.getLogger()

class LinkResolver(markdown.treeprocessors.Treeprocessor):
    def run(self, root):
        for a in root.iter("a"):
            new_url = self.resolve_url(a.attrib.get("href", None))
            if new_url is not None:
                a.attrib["href"] = new_url

        for a in root.iter("img"):
            new_url = self.resolve_url(a.attrib.get("src", None))
            if new_url is not None:
                a.attrib["src"] = new_url

    def resolve_url(self, url):
        """
        Resolve internal URLs.

        Returns None if the URL does not need changing, else returns the new URL.
        """
        from markdown.util import AMP_SUBSTITUTE
        if not url:
            return None
        if url.startswith(AMP_SUBSTITUTE):
            # Possibly an overencoded mailto: link.
            # see https://bugs.debian.org/816218
            #
            # Markdown then further escapes & with utils.AMP_SUBSTITUTE, so
            # we look for it here.
            return None
        parsed = urlparse(url)
        if parsed.scheme or parsed.netloc:
            return None
        if not parsed.path:
            return None
        dest = self.page.resolve_link(parsed.path)
        # Also allow .md extension in
        if dest is None and parsed.path.endswith(".md"):
            dirname, basename = os.path.split(parsed.path)
            if basename == "index.md":
                dest = self.page.resolve_link(dirname)
            else:
                dest = self.page.resolve_link(parsed.path[:-3])
        if dest is None:
            log.warn("%s: internal link %r does not resolve to any site page", self.page.src_relpath, url)
            return None

        return urlunparse(
            (parsed.scheme, parsed.netloc, dest.dst_link, parsed.params, parsed.query, parsed.fragment)
        )


class StaticSiteExtension(markdown.extensions.Extension):
    def extendMarkdown(self, md, md_globals):
        self.link_resolver = LinkResolver(md)
        # Insert instance of 'mypattern' before 'references' pattern
        md.treeprocessors.add('staticsite', self.link_resolver, '_end')
        md.registerExtension(self)

    def reset(self):
        pass

    def set_page(self, page):
        self.page = page
        self.link_resolver.page = page


class MarkdownPages:
    def __init__(self, site):
        self.site = site
        self.md_staticsite = StaticSiteExtension()
        self.markdown = markdown.Markdown(
            extensions=[
                "markdown.extensions.extra",
                "markdown.extensions.codehilite",
                "markdown.extensions.fenced_code",
                self.md_staticsite,
            ],
            output_format="html5",
        )
        # Cached templates
        self._page_template = None
        self._redirect_template = None

    @property
    def page_template(self):
        if not self._page_template:
            self._page_template = self.site.theme.jinja2.get_template("page.html")
        return self._page_template

    @property
    def redirect_template(self):
        if not self._redirect_template:
            self._redirect_template = self.site.theme.jinja2.get_template("redirect.html")
        return self._redirect_template

    def render(self, page):
        self.md_staticsite.set_page(page)
        self.markdown.reset()
        return self.markdown.convert(page.get_content())

    def try_load_page(self, root_abspath, relpath):
        if not relpath.endswith(".md"): return None
        return MarkdownPage(self, root_abspath, relpath)

    def try_load_archetype(self, archetypes, relpath, name):
        if not relpath.endswith(".md"): return None
        if not (relpath.endswith(name) or relpath.endswith(name + ".md")): return None
        return MarkdownArchetype(self, archetypes, relpath)


def parse_markdown_with_front_matter(fd):
    """
    Parse lines of markdown with front matter.

    Returns two arrays: one with the lines of front matter and one with the
    lines of markdown.
    """
    front_matter = []
    body = []

    front_matter_end = None
    in_front_matter = True

    for lineno, line in enumerate(fd, 1):
        line = line.rstrip()
        if lineno == 1:
            if line == "{":
                front_matter_end = "}"
            elif line == "---":
                front_matter_end = "---"
            elif line == "+++":
                front_matter_end = "+++"
            else:
                in_front_matter = False

        if in_front_matter:
            front_matter.append(line)
            if lineno > 1 and line == front_matter_end:
                in_front_matter = False
        else:
            body.append(line)

    return front_matter, body


class MarkdownArchetype(Archetype):
    def __init__(self, mdenv, archetypes, relpath):
        super().__init__(mdenv.site, relpath)
        self.mdenv = mdenv
        self.archetypes = archetypes

    def render(self, **kw):
        """
        Process the archetype returning its parsed front matter in a dict, and
        its contents in a string
        """
        # Render the archetype with jinja2
        abspath = os.path.join(self.archetypes.root, self.relpath)
        with open(abspath, "rt") as fd:
            template = self.site.theme.jinja2.from_string(fd.read())

        rendered = template.render(**kw)

        with io.StringIO(rendered) as fd:
            # Reparse it separating front matter and markdown content
            front_matter, body = parse_markdown_with_front_matter(fd)

            try:
                style, meta = parse_front_matter(front_matter)
            except:
                log.exception("archetype %s: failed to parse front matter", self.relpath)

        return style, meta, body


class MarkdownPage(Page):
    TYPE = "markdown"

    FINDABLE = True

    def __init__(self, mdenv, root_abspath, relpath):
        dirname, basename = os.path.split(relpath)
        if basename == "index.md":
            linkpath = dirname
        else:
            linkpath = os.path.splitext(relpath)[0]
        super().__init__(
            site=mdenv.site,
            root_abspath=root_abspath,
            src_relpath=relpath,
            src_linkpath=linkpath,
            dst_relpath=os.path.join(linkpath, "index.html"),
            dst_link=os.path.join(settings.SITE_ROOT, linkpath))

        # Shared markdown environment
        self.mdenv = mdenv

        # Sequence of lines found in the front matter
        self.front_matter = []

        # Sequence of lines found in the body
        self.body = []

        # Markdown content of the page rendered into html
        self.md_html = None

        # Read the contents
        src = self.src_abspath
        if self.meta.get("date", None) is None:
            self.meta["date"] = pytz.utc.localize(datetime.datetime.utcfromtimestamp(os.path.getmtime(src)))

        # Parse separating front matter and markdown content
        with open(src, "rt") as fd:
            self.front_matter, self.body = parse_markdown_with_front_matter(fd)

        try:
            style, meta = parse_front_matter(self.front_matter)
            self.meta.update(**meta)
        except:
            log.exception("%s: failed to parse front matter", self.src_relpath)

        # Remove leading empty lines
        while self.body and not self.body[0]:
            self.body.pop(0)

        # Read title from first # title if not specified in metadata
        if not self.meta.get("title", ""):
            if self.body and self.body[0].startswith("# "):
                self.meta["title"] = self.body.pop(0)[2:].strip()

                # Remove leading empty lines again
                while self.body and not self.body[0]:
                    self.body.pop(0)

        date = self.meta.get("date", None)
        if date is not None and not isinstance(date, datetime.datetime):
            self.meta["date"] = dateutil.parser.parse(date)

    def get_content(self):
        return "\n".join(self.body)

    def check(self, checker):
        self.mdenv.render(self)

    @property
    def content(self):
        if self.md_html is None:
            self.md_html = self.mdenv.render(self)
        return self.md_html

    def render(self):
        res = {}

        html = self.mdenv.page_template.render(
            page=self,
            content=self.content,
            **self.meta
        )
        res[self.dst_relpath] = RenderedString(html)

        for relpath in self.meta.get("aliases", ()):
            html = self.mdenv.redirect_template.render(
                page=self,
            )
            res[os.path.join(relpath, "index.html")] = RenderedString(html)

        return res

    def target_relpaths(self):
        res = [self.dst_relpath]
        for relpath in self.meta.get("aliases", ()):
            res.append(os.path.join(relpath, "index.html"))
        return res



# Name of the website
SITE_NAME = "Example web site"

# Author of the website
SITE_AUTHOR = "Example author"

# Absolute URL to the root of the website
SITE_URL = "https://www.example.org"

# Time zone used for site posts
TIMEZONE = "Europe/Rome"

# Root directory of the site in the URLs we generate.
# If you are publishing the site at /prefix instead of root of the domain,
# override this with /prefix
# SITE_ROOT = "/"

# Directory with the source content of the site
# CONTENT = "content"

# Theme used to render the content
# Default: the one installed in the system
THEME = "theme"

# Directory with "archetypes" (templates used by ssite new)
#ARCHETYPES = "archetypes"

# Directory where the static site will be written by build
#OUTPUT = "web"



# coding: utf-8

from .core import Archetype, Page, RenderedString, settings
import os
import re
from collections import defaultdict
import logging

log = logging.getLogger()


class IgnorePage(Exception):
    pass


class J2Pages:
    def __init__(self, site):
        self.site = site

    def try_load_page(self, root_abspath, relpath):
        basename = os.path.basename(relpath)
        if ".j2." not in basename: return None
        try:
            return J2Page(self, root_abspath, relpath)
        except IgnorePage:
            return None


class J2Page(Page):
    TYPE = "jinja2"

    RENDER_PREFERRED_ORDER = 2

    def __init__(self, j2env, root_abspath, relpath):
        dirname, basename = os.path.split(relpath)
        dst_basename = basename.replace(".j2", "")
        dst_relpath = os.path.join(dirname, dst_basename)

        if dst_basename == "index.html":
            linkpath = dirname
        else:
            linkpath = os.path.join(dirname, dst_basename)

        super().__init__(
            site=j2env.site,
            root_abspath=root_abspath,
            src_relpath=relpath,
            src_linkpath=linkpath,
            dst_relpath=dst_relpath,
            dst_link=os.path.join(settings.SITE_ROOT, linkpath))

    def read_metadata(self):
        self.meta["date"] = self.site.generation_time

    def render(self):
        with open(self.src_abspath, "rt") as fd:
            template_body = fd.read()
        try:
            template = self.site.theme.jinja2.from_string(template_body)
        except:
            log.exception("%s: cannot load template", self.src_relpath)
            raise IgnorePage
        body = template.render(
            page=self,
        )
        return {
            self.dst_relpath: RenderedString(body),
        }



# coding: utf-8
from unittest import TestCase
from staticsite.site import Site
from staticsite.core import Page
from . import datafile_abspath
import os
import datetime

class TestPage(Page):
    TYPE = "test"
    FINDABLE = True

    def __init__(self, site, relpath, dt):
        super().__init__(
            site=site,
            root_abspath="/",
            src_relpath=relpath,
            src_linkpath=relpath,
            dst_relpath=relpath,
            dst_link=relpath)
        self.dt = dt

    def read_metadata(self):
        self.meta["date"] = self.dt


class TestSite(TestCase):
    def test_dirs(self):
        site = Site()

        page_root = TestPage(site, "page_root", datetime.datetime(2016, 1, 1))
        page_sub = TestPage(site, "dir1/page_sub", datetime.datetime(2016, 2, 1))
        page_sub3 = TestPage(site, "dir1/dir2/dir3/page_sub3", datetime.datetime(2016, 3, 1))

        site.add_page(page_root)
        site.add_page(page_sub)
        site.add_page(page_sub3)
        site.load_theme(datafile_abspath("theme"))
        site.analyze()

        # We have a root dir index and dir indices for all subdirs
        dir_root = site.pages[""]
        dir_dir1 = site.pages["dir1"]
        dir_dir2 = site.pages["dir1/dir2"]
        dir_dir3 = site.pages["dir1/dir2/dir3"]

        self.assertEquals(dir_root.TYPE, "dir")
        self.assertEquals(dir_dir1.TYPE, "dir")
        self.assertEquals(dir_dir3.TYPE, "dir")
        self.assertEquals(dir_dir3.TYPE, "dir")

        # Check the contents of all dirs
        self.assertEquals(dir_root.pages, [page_root])
        self.assertEquals(dir_root.subdirs, [dir_dir1])
        self.assertEquals(dir_dir1.pages, [page_sub])
        self.assertEquals(dir_dir1.subdirs, [dir_dir2])
        self.assertEquals(dir_dir2.pages, [])
        self.assertEquals(dir_dir2.subdirs, [dir_dir3])
        self.assertEquals(dir_dir3.pages, [page_sub3])
        self.assertEquals(dir_dir3.subdirs, [])



# coding: utf-8

from .commands import SiteCommand, CmdlineError
from staticsite.core import settings, PageFS
import os
import mimetypes
import gc
import logging

log = logging.getLogger()

class Serve(SiteCommand):
    "serve the site over HTTP, building it in memory on demand"

    def run(self):
        mimetypes.init()

        self.reload()

        from livereload import Server
        server = Server(self.application)
        server.watch(self.content_root, self.reload)
        server.watch(self.theme_root, self.reload)
        server.serve(port=8000, host="localhost")

    def application(self, environ, start_response):
        path = environ.get("PATH_INFO", None)
        if path is None:
            start_response("404 not found", [("Content-Type", "text/plain")])
            return [b"Not found"]

        content = self.pages.serve_path(path, environ, start_response)
        if content is not None:
            return content

        start_response("404 not found", [("Content-Type", "text/plain")])
        return [b"Not found"]

    def reload(self):
        log.info("Loading site")
        self.site = self.load_site()
        gc.collect()

        self.pages = PageFS()
        self.pages.add_site(self.site)



# coding: utf-8

from .core import settings
from .site import Site
from .utils import timings
import sys
import os
import logging

log = logging.getLogger()

class CmdlineError(RuntimeError):
    pass


class SiteCommand:
    # Command name (as used in command line)
    # Defaults to the lowercased class name
    NAME = None

    # Command description (as used in command line help)
    # Defaults to the strip()ped class docstring.
    DESC = None

    def __init__(self, args):
        self.args = args

        self.setup_logging(args)

        # Default to current directory if project was not provided.
        # If the project was provided and is a .py file, load it as settings.
        if args.project:
            if os.path.isfile(args.project) and args.project.endswith(".py"):
                self.settings_abspath = os.path.abspath(args.project)
                self.root = os.path.dirname(self.settings_abspath)
            else:
                self.root = os.path.abspath(args.project)
                self.settings_abspath = os.path.join(self.root, "settings.py")
        else:
            self.root = os.getcwd()
            self.settings_abspath = os.path.join(self.root, "settings.py")

        # Load settings (optional)
        if os.path.isfile(self.settings_abspath):
            settings.load(self.settings_abspath)

        # Command line overrides for settings
        if self.args.theme: settings.THEME = os.path.abspath(self.args.theme)
        if self.args.content: settings.CONTENT = os.path.abspath(self.args.content)
        if self.args.archetypes: settings.ARCHETYPES = os.path.abspath(self.args.archetypes)
        if self.args.output: settings.OUTPUT = os.path.abspath(self.args.output)

        # Double check that root points to something that looks like a project
        self.content_root = os.path.join(self.root, settings.CONTENT)
        if not os.path.exists(self.content_root):
            raise CmdlineError("Content directory {} does not exist".format(self.content_root))

        self.theme_root = os.path.join(self.root, settings.THEME)
        if not os.path.exists(self.theme_root):
            raise CmdlineError("Theme directory {} does not exist".format(self.theme_root))

    def setup_logging(self, args):
        FORMAT = "%(asctime)-15s %(levelname)s %(message)s"
        if args.debug:
            logging.basicConfig(level=logging.DEBUG, stream=sys.stderr, format=FORMAT)
        elif args.verbose:
            logging.basicConfig(level=logging.INFO, stream=sys.stderr, format=FORMAT)
        else:
            logging.basicConfig(level=logging.WARN, stream=sys.stderr, format=FORMAT)

    def load_site(self):
        # Instantiate site
        site = Site()
        site.draft = self.args.draft

        # Read and analyze site contents
        with timings("Read site in %fs"):
            site.load_theme(self.theme_root)
            site.load_content(self.content_root)

        with timings("Analysed site tree in %fs"):
            site.analyze()

        return site

    @classmethod
    def make_subparser(cls, subparsers):
        name = cls.NAME
        if name is None:
            name = cls.__name__.lower()

        desc = cls.DESC
        if desc is None:
            desc = cls.__doc__.strip()

        parser = subparsers.add_parser(name, help=desc)
        parser.add_argument("project", nargs="?", help="project directory or .py configuration file (default: the current directory)")
        parser.add_argument("--theme", help="theme directory location. Overrides settings.THEME")
        parser.add_argument("--content", help="content directory location. Overrides settings.CONTENT")
        parser.add_argument("--archetypes", help="archetypes directory location. Override settings.ARCHETYPES")
        parser.add_argument("-o", "--output", help="output directory location. Override settings.OUTPUT")
        parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
        parser.add_argument("--draft", action="store_true", help="do not ignore pages with date in the future")
        parser.add_argument("--debug", action="store_true", help="verbose output")
        parser.set_defaults(handler=cls)

        return parser



# coding: utf-8

import os
import shutil
import json
import datetime
import time
import subprocess
import shlex
from collections import OrderedDict
from .commands import SiteCommand, CmdlineError
from .core import settings
import logging

log = logging.getLogger()

class Edit(SiteCommand):
    "edit an existing page. Bare keywords match titles and file names, '+tag' matches tags"

    MENU_SIZE = 5

    def match_page(self, page):
        for m in self.args.match:
            if m.startswith("+"):
                tags = page.meta.get("tags", ())
                if m[1:] not in tags:
                    return False
            elif (m.lower() not in page.meta.get("title", "").lower()
                  and m.lower() not in page.src_relpath.lower()):
                return False
        return True

    def select_page_menu(self, pages):
        """
        Prompt the user to select a page from `pages`.

        Returns None of the user chooses to give up insteaad.
        """
        first = 0

        while True:
            pagination = len(pages) > self.MENU_SIZE

            if pagination:
                print("{count} pages found, showing {first}-{last}:".format(
                    first=first + 1,
                    last=min(first + self.MENU_SIZE, len(pages)),
                    count=len(pages)))
                prompt = "Please choose which page to edit (number, or p for previous page, n for next page, q to quit): "
            else:
                print("{count} matching pages found:".format(count=len(pages)))
                prompt = "Please choose which page to edit (number, or q to quit): "
            print()

            for idx, page in enumerate(pages[first:first + self.MENU_SIZE], first + 1):
                print("{idx}: {date} {relpath} {title}".format(
                    idx=idx,
                    date=page.meta["date"].strftime("%Y-%m-%d"),
                    relpath=page.src_relpath,
                    title=page.meta.get("title", "{no title}")))
            print()

            selection = input(prompt).strip().lower()
            if selection == "q":
                return None
            elif pagination and selection == "p":
                first = max(0, first - self.MENU_SIZE)
                continue
            elif pagination and selection == "n":
                first = min(first + self.MENU_SIZE, self.MENU_SIZE * (len(pages) // self.MENU_SIZE))
                continue

            if selection.isdigit():
                idx = int(selection)
                if idx < 1 or idx > len(pages):
                    print("{idx} does not match any page".format(idx=idx))
                else:
                    return pages[idx - 1]

    def run(self):
        site = self.load_site()

        # Build a list of all findable pages sorted with the newest first
        pages = [p for p in site.pages.values() if p.FINDABLE]
        pages.sort(key=lambda x:x.meta["date"], reverse=True)

        selected = []
        for page in pages:
            if not self.match_page(page): continue
            selected.append(page)

        if len(selected) == 0:
            raise CmdlineError("No page found matching {}".format(" ".join(shlex.quote(x) for x in self.args.match)))

        if len(selected) == 1:
            page = selected[0]
        else:
            page = self.select_page_menu(selected)

        if page is None:
            return

        abspath = page.src_abspath

        if not self.args.noedit:
            settings_dict = settings.as_dict()
            cmd = [x.format(name=abspath, **settings_dict) for x in settings.EDIT_COMMAND]
            try:
                subprocess.check_call(cmd)
            except subprocess.CalledProcessError as e:
                log.warn("Editor command %s exited with error %d", " ".join(shlex.quote(x) for x in cmd), e.returncode)
                return e.returncode
        print(abspath)

    @classmethod
    def make_subparser(cls, subparsers):
        parser = super().make_subparser(subparsers)
        parser.add_argument("match", nargs="*", help="keywords used to look for the page to edit")
        parser.add_argument("-n", "--noedit", action="store_true", help="do not run an editor, only output the file name of the new post")
        return parser



# coding: utf-8

import os
import shutil
import json
import datetime
import time
import subprocess
import shlex
from collections import OrderedDict
from .commands import SiteCommand, CmdlineError
from .core import settings
from .archetypes import Archetypes
import logging

log = logging.getLogger()

class LazyTitle:
    """
    Wrapper for the value of a post title.

    If no value has been provided and one is requested by the template, ask the
    user for one.
    """
    def __init__(self, title=None):
        self.title = title

    def __str__(self):
        if self.title is None:
            self.title = input("Please enter the post title: ")
        return self.title


class LazySlug:
    """
    Generate a slug from the value of LazyTitle
    """
    def __init__(self, site, lazy_title):
        self.site = site
        self.lazy_title = lazy_title

    def __str__(self):
        return self.site.slugify(str(self.lazy_title))


class New(SiteCommand):
    "create a new page"

    def run(self):
        site = self.load_site()

        archetypes = Archetypes(site, os.path.join(self.root, "archetypes"))

        archetype = archetypes.find(self.args.archetype)
        if archetype is None:
            raise CmdlineError("archetype {} not found".format(self.args.archetype))
        log.info("Using archetype %s", archetype.relpath)

        title = LazyTitle(self.args.title)
        slug = LazySlug(site, title)
        meta_style, meta, body = archetype.render(slug=slug, title=title)

        relpath = meta.pop("path", None)
        if relpath is None:
            raise CmdlineError("archetype {} does not contain `path` in its front matter".format(archetype.relpath))

        abspath = os.path.join(self.content_root, relpath)

        if self.args.overwrite or not os.path.exists(abspath):
            from .utils import write_front_matter
            front_matter = write_front_matter(meta, meta_style)

            os.makedirs(os.path.dirname(abspath), exist_ok=True)

            with open(abspath, "wt") as out:
                out.write(front_matter)
                print(file=out)
                for line in body:
                    print(line, file=out)
        else:
            log.info("%s already exists: reusing it", abspath)

        if not self.args.noedit:
            settings_dict = settings.as_dict()
            cmd = [x.format(name=abspath, slug=slug, **settings_dict) for x in settings.EDIT_COMMAND]
            try:
                subprocess.check_call(cmd)
            except subprocess.CalledProcessError as e:
                log.warn("Editor command %s exited with error %d", " ".join(shlex.quote(x) for x in cmd), e.returncode)
                return e.returncode
        print(abspath)

    @classmethod
    def make_subparser(cls, subparsers):
        parser = super().make_subparser(subparsers)
        parser.add_argument("-a", "--archetype", default="default", help="page archetype")
        parser.add_argument("-t", "--title", help="page title")
        parser.add_argument("-n", "--noedit", action="store_true", help="do not run an editor, only output the file name of the new post")
        parser.add_argument("--overwrite", action="store_true", help="if a post already exists, overwrite it instead of reusing it")
        return parser



# coding: utf-8
import pytz
import contextlib
import time
import logging

log = logging.getLogger()

def parse_front_matter(lines):
    """
    Parse lines of front matter
    """
    if not lines: return "toml", {}

    if lines[0] == "{":
        # JSON
        import json
        return "json", json.loads("\n".join(lines))

    if lines[0] == "+++":
        # TOML
        import toml
        return "toml", toml.loads("\n".join(lines[1:-1]))

    if lines[0] == "---":
        # YAML
        import yaml
        return "yaml", yaml.load("\n".join(lines[1:-1]), Loader=yaml.CLoader)

    return {}

def write_front_matter(meta, style="toml"):
    if style == "json":
        import json
        return json.dumps(meta, indent=4, sort_keys=True)
    elif style == "toml":
        import toml
        return "+++\n" + toml.dumps(meta) + "+++\n"
    elif style == "yaml":
        import yaml
        return "---\n" + yaml.dump(meta) + "---\n"
    return ""

def format_date_rfc822(dt):
    from email.utils import formatdate
    return formatdate(dt.timestamp())

def format_date_rfc3339(dt):
    dt = dt.astimezone(pytz.utc)
    return dt.strftime("%Y-%m-%dT%H:%M:%SZ")

def format_date_w3cdtf(dt):
    from dateutil.tz import tzlocal
    offset = dt.utcoffset()
    offset_sec = (offset.days * 24 * 3600 + offset.seconds)
    offset_hrs = offset_sec // 3600
    offset_min = offset_sec % 3600
    if offset:
        tz_str = '{0:+03d}:{1:02d}'.format(offset_hrs, offset_min // 60)
    else:
        tz_str = 'Z'
    return dt.strftime("%Y-%m-%dT%H:%M:%S") + tz_str

def format_date_iso8601(dt):
    from dateutil.tz import tzlocal
    offset = dt.utcoffset()
    offset_sec = (offset.days * 24 * 3600 + offset.seconds)
    offset_hrs = offset_sec // 3600
    offset_min = offset_sec % 3600
    if offset:
        tz_str = '{0:+03d}:{1:02d}'.format(offset_hrs, offset_min // 60)
    else:
        tz_str = 'Z'
    return dt.strftime("%Y-%m-%d %H:%M:%S") + tz_str

@contextlib.contextmanager
def timings(fmtstr, *args, **kw):
    """
    Times the running of a command, and writes a log entry afterwards.

    The log entry is passed an extra command at the beginning with the elapsed
    time in floating point seconds.
    """
    start = time.perf_counter()
    yield
    end = time.perf_counter()
    log.info(fmtstr, end - start, *args, extra=kw)




# coding: utf-8
from unittest import TestCase
from staticsite.core import Settings
from staticsite import global_settings
from . import datafile_abspath
import os

class TestSettings(TestCase):
    def test_defaults(self):
        s = Settings()
        self.assertEquals(s.TIMEZONE, global_settings.TIMEZONE)

    def test_as_dict(self):
        s = Settings()
        d = s.as_dict()
        self.assertEquals(d["TIMEZONE"], global_settings.TIMEZONE)

    def test_add_module(self):
        s = Settings()
        self.assertFalse(hasattr(s, "PRIO_USER"))

        s.add_module(os)

        # Merging a module add all uppercase symbols
        self.assertEquals(s.PRIO_USER, os.PRIO_USER)

        # And ignores all lowercase ones
        self.assertFalse(hasattr(s, "getuid"))

    def test_load(self):
        s = Settings()
        self.assertFalse(hasattr(s, "UPPERCASE"))

        s.load(datafile_abspath("settings.py"))

        self.assertEquals(s.UPPERCASE, True)
        self.assertFalse(hasattr(s, "lowercase"))
        self.assertFalse(hasattr(s, "os"))



# coding: utf-8

from .core import Page, RenderedFile, settings
import datetime
import os
import pytz
import shutil


class Asset(Page):
    TYPE = "asset"

    def __init__(self, site, root_abspath, relpath):
        dirname, basename = os.path.split(relpath)
        if basename == "index.html":
            linkpath = dirname
        else:
            linkpath = relpath

        super().__init__(
            site=site,
            root_abspath=root_abspath,
            src_relpath=relpath,
            src_linkpath=linkpath,
            dst_relpath=relpath,
            dst_link=os.path.join(settings.SITE_ROOT, linkpath))
        self.title = os.path.basename(relpath)

        dt = pytz.utc.localize(datetime.datetime.fromtimestamp(os.path.getmtime(self.src_abspath)))
        self.meta["date"] = dt

    def render(self):
        return {
            self.dst_relpath: RenderedFile(self.src_abspath),
        }



# coding: utf-8

import os
import shutil
import json
import datetime
from .utils import timings
from collections import Counter
from .commands import SiteCommand, CmdlineError
import logging

log = logging.getLogger()

class Check(SiteCommand):
    "check the site, going through all the motions of rendering it without writing anything"

    def run(self):
        site = self.load_site()
        with timings("Checked site in %fs"):
            self.check(site)

    def check(self, site):
        counts = Counter()
        for page in site.pages.values():
            counts[page.TYPE] += 1
            page.check(self)

            date = page.meta.get("date", None)
            if date is not None:
                if not isinstance(date, datetime.datetime):
                    log.info("%s: meta date %r is not an instance of datetime", page.src_relpath, date)

        for type, count in sorted(counts.items()):
            print("{} {} pages".format(count, type))



# coding: utf-8

import json
import os
import re
import time
import shutil
from collections import defaultdict
from .commands import SiteCommand, CmdlineError
from .core import settings
from .utils import timings
import logging

log = logging.getLogger()

class Build(SiteCommand):
    "build the site into the web/ directory of the project"

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.output_root = os.path.join(self.root, settings.OUTPUT)

    def run(self):
        site = self.load_site()
        self.write(site)

    def clear_outdir(self, outdir):
        for f in os.listdir(outdir):
            abs = os.path.join(outdir, f)
            if os.path.isdir(abs):
                shutil.rmtree(abs)
            else:
                os.unlink(abs)

    def write(self, site):
        """
        Generate output
        """
        # Clear the target directory, but keep the root path so that a web
        # server running on it does not find itself running nowhere
        if os.path.exists(self.output_root):
            with timings("Cleaned output directory in %fs"):
                self.clear_outdir(self.output_root)

        with timings("Built site in %fs"):
            # cpu_count = os.cpu_count()
            # if cpu_count > 1:
            #     self.write_multi_process(site, cpu_count)
            # else:
            #     self.write_single_process(site)

            # I tried trivial parallelisation with child processes but the benefits
            # do not seem significant:
            #
            #     real  0m5.111s
            #     user  0m8.096s
            #     sys   0m0.644s
            #
            # compared with:
            #
            #     real  0m6.251s
            #     user  0m5.760s
            #     sys   0m0.468s
            self.write_single_process(site)

    def write_multi_process(self, site, child_count):
        log.info("Generating pages using %d child processes", child_count)

        pages = list(site.pages.values())

        # From http://code.activestate.com/recipes/576785-partition-an-iterable-into-n-lists/
        chunks = [pages[i::child_count] for i in range(child_count)]

        print(len(pages))
        for c in chunks:
            print(len(c))

        import sys
        pids = set()
        for chunk in chunks:
            pid = os.fork()
            if pid == 0:
                self.write_pages(chunk)
                sys.exit(0)
            else:
                pids.add(pid)

        while pids:
            (pid, status) = os.wait()
            pids.discard(pid)

    def write_single_process(self, site):
        sums, counts = self.write_pages(site.pages.values())
        for type in sorted(sums.keys()):
            log.info("%s: %d in %.3fs (%.1f per minute)", type, counts[type], sums[type], counts[type]/sums[type] * 60)

    def write_pages(self, pages):
        sums = defaultdict(float)
        counts = defaultdict(int)

        # group by type
        by_type = defaultdict(list)
        for page in pages:
            by_type[(page.RENDER_PREFERRED_ORDER, page.TYPE)].append(page)

        # Render collecting timing statistics
        for (order, type), pgs in sorted(by_type.items(), key=lambda x:x[0][0]):
            start = time.perf_counter()
            for page in pgs:
                contents = page.render()
                for relpath, rendered in contents.items():
                    dst = self.output_abspath(relpath)
                    rendered.write(dst)
            end = time.perf_counter()
            sums[type] = end - start
            counts[type] = len(pgs)
        return sums, counts

    def output_abspath(self, relpath):
        abspath = os.path.join(self.output_root, relpath)
        os.makedirs(os.path.dirname(abspath), exist_ok=True)
        return abspath



# coding: utf-8
import os
import re
import pytz
import datetime
from collections import defaultdict
from .core import settings
import logging

log = logging.getLogger()

class Site:
    def __init__(self):
        # Site pages
        self.pages = {}

        # Description of tags
        self.tag_descriptions = {}

        # Site time zone
        self.timezone = pytz.timezone(settings.TIMEZONE)

        # Current datetime
        self.generation_time = pytz.utc.localize(datetime.datetime.utcnow()).astimezone(self.timezone)

        # Taxonomies found in the site
        self.taxonomies = []

        # Theme used to render pages
        self.theme = None

        # If true, do not ignore pages with dates in the future
        self.draft = False

        # Map input file patterns to resource handlers
        from .markdown import MarkdownPages
        from .j2 import J2Pages
        from .taxonomy import TaxonomyPages
        self.page_handlers = [
            MarkdownPages(self),
            J2Pages(self),
            TaxonomyPages(self),
        ]

    def load_theme(self, theme_root):
        """
        Load a theme from the given directory.

        This needs to be called once (and only once) before analyze() is
        called.
        """
        if self.theme is not None:
            raise RuntimeError("cannot load theme from {} because it was already loaded from {}".format(theme_root, self.theme.root))

        from .theme import Theme
        self.theme = Theme(self, theme_root)

        theme_static = os.path.join(theme_root, "static")
        if os.path.isdir(theme_static):
            self.read_asset_tree(theme_static)

    def load_content(self, content_root):
        self.read_contents_tree(content_root)

    def add_page(self, page):
        ts = page.meta.get("date", None)
        if not self.draft and ts is not None and ts > self.generation_time:
            log.info("Ignoring page %s with date %s in the future", page.src_relpath, ts - self.generation_time)
            return
        self.pages[page.src_linkpath] = page

    def read_contents_tree(self, tree_root):
        """
        Read static assets and pages from a directory and all its subdirectories
        """
        from .asset import Asset

        log.info("Loading pages from %s", tree_root)

        for root, dnames, fnames in os.walk(tree_root):
            for f in fnames:
                if f.startswith("."): continue

                page_abspath = os.path.join(root, f)
                page_relpath = os.path.relpath(page_abspath, tree_root)

                for handler in self.page_handlers:
                    p = handler.try_load_page(tree_root, page_relpath)
                    if p is not None:
                        self.add_page(p)
                        break
                else:
                    if os.path.isfile(page_abspath):
                        log.debug("Loading static file %s", page_relpath)
                        p = Asset(self, tree_root, page_relpath)
                        self.add_page(p)

    def read_asset_tree(self, tree_root):
        """
        Read static assets from a directory and all its subdirectories
        """
        from .asset import Asset

        log.info("Loading assets from %s", tree_root)

        for root, dnames, fnames in os.walk(tree_root):
            for f in fnames:
                if f.startswith("."): continue

                page_abspath = os.path.join(root, f)
                if os.path.isfile(page_abspath):
                    page_relpath = os.path.relpath(page_abspath, tree_root)
                    log.debug("Loading static file %s", page_relpath)
                    p = Asset(self, tree_root, page_relpath)
                    self.add_page(p)

    def relocate(self, page, dest_relpath):
        log.info("Relocating %s to %s", page.relpath, dest_relpath)
        if dest_relpath in self.pages:
            log.warn("Cannot relocate %s to existing page %s", page.relpath, dest_relpath)
            return
        self.pages[dest_relpath] = page
        page.aliases.append(page.relpath)
        page.relpath = dest_relpath

    def analyze(self):
        self.taxonomies = []

        by_dir = defaultdict(list)
        by_pass = defaultdict(list)
        for page in self.pages.values():
            # Group pages by pass number
            by_pass[page.ANALYZE_PASS].append(page)
            # Collect taxonomies
            if page.TYPE == "taxonomy":
                self.taxonomies.append(page)
            # Harvest content for directory indices
            if page.FINDABLE and page.src_relpath:
                dir_relpath = os.path.dirname(page.src_relpath)
                by_dir[dir_relpath].append(page)
                while True:
                    if not dir_relpath: break
                    dir_relpath = os.path.dirname(dir_relpath)
                    # Do a lookup to make sure an entry exists for this
                    # directory level, even though without pages
                    by_dir[dir_relpath]


        # Build directory indices
        from .dir import DirPage
        for relpath, pages in by_dir.items():
            # We only build indices where there is not already a page
            if relpath in self.pages: continue
            page = DirPage(self, relpath, pages)
            self.pages[relpath] = page
            by_pass[page.ANALYZE_PASS].append(page)

        # Add directory indices to their parent directory indices
        for relpath, pages in by_dir.items():
            page = self.pages[relpath]
            if page.TYPE != "dir": continue
            page.attach_to_parent()

        # Read metadata
        for passnum, pages in sorted(by_pass.items(), key=lambda x:x[0]):
            for page in pages:
                page.read_metadata()


    def slugify(self, text):
        from slugify import slugify
        return slugify(text)



# coding: utf-8

from .core import Page, settings, RenderedString
import os
import re
from collections import defaultdict
import jinja2
import logging

log = logging.getLogger()

class TaxonomyPages:
    def __init__(self, site):
        self.site = site

    def try_load_page(self, root_abspath, relpath):
        if not relpath.endswith(".taxonomy"): return None
        return TaxonomyPage(self, root_abspath, relpath)


class TaxonomyItem:
    def __init__(self, page, name):
        self.page = page
        self.name = name
        self.slug = self.page.site.slugify(name)
        self.pages = []

    def __str__(self):
        return self.name


class TaxonomyPage(Page):
    TYPE = "taxonomy"
    ANALYZE_PASS = 2
    RENDER_PREFERRED_ORDER = 2

    def __init__(self, tenv, root_abspath, relpath):
        linkpath = os.path.splitext(relpath)[0]

        super().__init__(
            site=tenv.site,
            root_abspath=root_abspath,
            src_relpath=relpath,
            src_linkpath=linkpath,
            dst_relpath=linkpath,
            dst_link=os.path.join(settings.SITE_ROOT, linkpath))

        # Taxonomy name (e.g. "tags")
        self.name = os.path.basename(linkpath)

        # Map all possible values for this taxonomy to the pages that reference
        # them
        self.items = {}

        self.template_index = None
        self.template_item = None
        self.template_feed = None
        self.template_archive = None

        self.meta["output_index"] = ""
        self.meta["output_item"] = "{slug}/"
        self.meta["output_rss"] = "{slug}/index.rss"
        self.meta["output_atom"] = "{slug}/index.atom"
        self.meta["output_archive"] = "{slug}/archive.html"

        ## Generate taxonomies from configuration
        #self.taxonomies = {}
        #for name, info in settings.TAXONOMIES.items():
        #    info["name"] = name
        #    output_dir = info.get("output_dir", None)
        #    if output_dir is not None:
        #        info["output_dir"] = self.enforce_relpath(output_dir)
        #    self.taxonomies[name] = Taxonomy(**info)


    def link_value(self, context, output_item, value):
        if isinstance(value, str):
            item = self.items.get(value, None)
        else:
            item = value

        if item is None:
            log.warn("%s+%s: %s not found in taxonomy %s", context.parent["page"], context.name, value, self.name)
            return ""
        dest = self.meta[output_item].format(slug=item.slug)
        return os.path.join(settings.SITE_ROOT, self.dst_relpath, dest)

    @jinja2.contextfunction
    def link_index(self, context):
        dest = self.meta["output_index"]
        if not dest:
            return os.path.join(settings.SITE_ROOT, self.dst_relpath)
        else:
            return os.path.join(settings.SITE_ROOT, self.dst_relpath, dest)

    @jinja2.contextfunction
    def link_item(self, context, value):
        return self.link_value(context, "output_item", value)

    @jinja2.contextfunction
    def link_rss(self, context, value):
        return self.link_value(context, "output_rss", value)

    @jinja2.contextfunction
    def link_atom(self, context, value):
        return self.link_value(context, "output_atom", value)

    @jinja2.contextfunction
    def link_archive(self, context, value):
        return self.link_value(context, "output_archive", value)

    def load_template_from_meta(self, name):
        template_name = self.meta.get(name, None)
        if template_name is None: return None
        try:
            return self.site.theme.jinja2.get_template(template_name)
        except:
            log.exception("%s: cannot load %s %s", self.src_relpath, name, template_name)
            return None

    def read_metadata(self):
        from .utils import parse_front_matter

        # Read taxonomy information
        src = self.src_abspath
        with open(src, "rt") as fd:
            lines = [x.rstrip() for x in fd]
        try:
            style, meta = parse_front_matter(lines)
            self.meta.update(**meta)
        except:
            log.exception("%s.taxonomy: cannot parse taxonomy information", self.src_relpath)

        single_name = self.meta.get("item_name", self.name)

        # Instantiate jinja2 templates
        self.template_index = self.load_template_from_meta("template_" + self.name)
        self.template_item = self.load_template_from_meta("template_" + single_name)
        self.template_rss = self.load_template_from_meta("template_rss")
        self.template_atom = self.load_template_from_meta("template_atom")
        self.template_archive = self.load_template_from_meta("template_archive")

        # Extend jinja2 with a function to link to elements of this taxonomy
        self.site.theme.jinja2.globals["url_for_" + self.name] = self.link_index
        self.site.theme.jinja2.globals["url_for_" + single_name] = self.link_item
        self.site.theme.jinja2.globals["url_for_" + single_name + "_rss"] = self.link_rss
        self.site.theme.jinja2.globals["url_for_" + single_name + "_atom"] = self.link_atom
        self.site.theme.jinja2.globals["url_for_" + single_name + "_archive"] = self.link_archive

        # Collect the pages annotated with this taxonomy
        for page in self.site.pages.values():
            if page.ANALYZE_PASS > 1 : continue
            vals = page.meta.get(self.name, None)
            if vals is None: continue
            for v in vals:
                item = self.items.get(v, None)
                if item is None:
                    item = TaxonomyItem(self, v)
                    self.items[v] = item
                item.pages.append(page)

    def render(self):
        res = {}

        single_name = self.meta.get("item_name", self.name)

        if self.template_index is not None:
            dest = os.path.join(self.dst_relpath, self.meta["output_index"])
            if dest.endswith("/"):
                dest += "index.html"
            kwargs = {
                "page": self,
                self.name: sorted(self.items.values(), key=lambda x: x.name),
            }
            kwargs.update(**self.meta)
            body = self.template_index.render(**kwargs)
            res[dest] = RenderedString(body)

        for item in self.items.values():
            for type in ("item", "rss", "atom", "archive"):
                template = getattr(self, "template_" + type, None)
                if template is None: continue

                dest = self.meta["output_" + type].format(slug=item.slug)
                if dest.endswith("/"):
                    dest += "index.html"

                kwargs = {
                    "page": self,
                    single_name: item,
                    "pages": sorted(item.pages, key=lambda x:x.meta["date"], reverse=True),
                }
                kwargs.update(**self.meta)
                body = template.render(**kwargs)
                res[os.path.join(self.dst_relpath, dest)] = RenderedString(body)

        return res

    def target_relpaths(self):
        res = []

        if self.template_index is not None:
            dest = os.path.join(self.dst_relpath, self.meta["output_index"])
            if dest.endswith("/"):
                dest += "index.html"
            res.append(dest)

        for item in self.items.values():
            for type in ("item", "rss", "atom", "archive"):
                template = getattr(self, "template_" + type, None)
                if template is None: continue

                dest = self.meta["output_" + type].format(slug=item.slug)
                if dest.endswith("/"):
                    dest += "index.html"

                res.append(os.path.join(self.dst_relpath, dest))

        return res



# Root directory of the site in the URLs we generate.
#
# If you are publishing the site at /prefix instead of root of the domain,
# override this with /prefix
SITE_ROOT = "/"

# Default site name. Override it with your site name.
SITE_NAME = "Site name not set"

# Directory with "archetypes" (templates used by ssite new)
ARCHETYPES = "archetypes"

# Directory with the source content of the site
# Default: the project directory itself
CONTENT = "content"

# Theme used to render the stie
# Default: the one installed in the system
THEME = "/usr/share/doc/staticsite/example/theme/"

# Directory where the static site will be written by build
OUTPUT = "web"

# Time zone used for timestamps on the site
TIMEZONE = "UTC"

# Editor used to edit new pages
import os
EDITOR = os.environ.get("EDITOR", "sensible-editor")

# Command used to run the editor, as passed to subprocess.check_command.
# Each list element is expanded with string.format. All settings are available
# for expansion, and {name} is the absolute path of the file to edit.
EDIT_COMMAND = ["{EDITOR}", "{name}", "+"]



# coding: utf-8
import logging

log = logging.getLogger()


class Base:
    def __init__(self, page, lineno):
        self.page = page
        self.lineno = lineno

    @property
    def is_blank(self):
        return False


class Line(Base):
    def __init__(self, page, lineno, line):
        super().__init__(page, lineno)
        self.line = line

    @property
    def is_blank(self):
        return not self.line or self.line.isspace()


class CodeBegin(Base):
    def __init__(self, page, lineno, lang):
        super().__init__(page, lineno)
        self.lang = lang


class CodeEnd(Base):
    pass


class IkiwikiMap(Base):
    def __init__(self, page, lineno, content):
        super().__init__(page, lineno)
        self.content = content


class Text(Base):
    def __init__(self, page, lineno, text):
        super().__init__(page, lineno)
        self.text = text

    @property
    def is_blank(self):
        return self.text.isspace()


class EOL(Base):
    @property
    def is_blank(self):
        return True


class InternalLink(Base):
    def __init__(self, page, lineno, text, target):
        super().__init__(page, lineno)
        self.text = text
        self.target = self.page.resolve_link(target)
        if self.target is None:
            log.warn("%s:%d: link %s cannot be resolved to a known resource",
                     self.page.relpath, self.lineno, target)

class InlineImage(Base):
    def __init__(self, page, lineno, fname, alt):
        super().__init__(page, lineno)
        self.text = alt
        self.target = self.page.resolve_link(fname)
        if self.target is None:
            log.warn("%s:%d: image %s cannot be resolved to a known resource",
                     self.page.relpath, self.lineno, fname)


class Directive(Base):
    def __init__(self, page, lineno, content):
        super().__init__(page, lineno)
        self.content = content




# coding: utf-8
import jinja2
import os
import re
import fnmatch
import datetime
from .core import settings
import logging

log = logging.getLogger()

class Theme:
    def __init__(self, site, root):
        self.site = site

        # Absolute path to the root of the theme directory
        self.root = os.path.abspath(root)

        # Jinja2 template engine
        from jinja2 import Environment, FileSystemLoader
        self.jinja2 = Environment(
            loader=FileSystemLoader([
                self.root,
            ]),
            autoescape=True,
        )

        # Add settings to jinja2 globals
        for x in dir(settings):
            if not x.isupper(): continue
            self.jinja2.globals[x] = getattr(settings, x)

        # Install site's functions into the jinja2 environment
        self.jinja2.globals.update(
            has_page=self.jinja2_has_page,
            url_for=self.jinja2_url_for,
            site_pages=self.jinja2_site_pages,
            now=self.site.generation_time,
            next_month=(self.site.generation_time.replace(day=1) + datetime.timedelta(days=40)).replace(day=1, hour=0, minute=0, second=0, microsecond=0),
            taxonomies=self.jinja2_taxonomies,
        )
        self.jinja2.filters["datetime_format"] = self.jinja2_datetime_format
        self.jinja2.filters["basename"] = self.jinja2_basename

        self.dir_template = self.jinja2.get_template("dir.html")

    def jinja2_taxonomies(self):
        return self.site.taxonomies

    def jinja2_basename(self, val):
        return os.path.basename(val)

    @jinja2.contextfilter
    def jinja2_datetime_format(self, context, dt, format=None):
        if format in ("rss2", "rfc822"):
            from .utils import format_date_rfc822
            return format_date_rfc822(dt)
        elif format in ("atom", "rfc3339"):
            from .utils import format_date_rfc3339
            return format_date_rfc3339(dt)
        elif format == "w3cdtf":
            from .utils import format_date_w3cdtf
            return format_date_w3cdtf(dt)
        elif format == "iso8601" or not format:
            from .utils import format_date_iso8601
            return format_date_iso8601(dt)
        else:
            log.warn("%s+%s: invalid datetime format %r requested", cur_page.src_relpath, context.name, format)
            return "(unknown datetime format {})".format(format)

    @jinja2.contextfunction
    def jinja2_has_page(self, context, arg):
        cur_page = context.parent["page"]
        page = cur_page.resolve_link(arg)
        return page is not None

    @jinja2.contextfunction
    def jinja2_url_for(self, context, arg):
        if isinstance(arg, str):
            cur_page = context.parent["page"]
            page = cur_page.resolve_link(arg)
            if page is None:
                log.warn("%s+%s: unresolved link %s passed to url_for", cur_page.src_relpath, context.name, arg)
                return ""
        else:
            page = arg
        return page.dst_link

    @jinja2.contextfunction
    def jinja2_site_pages(self, context, path=None, limit=None, sort="-date"):
        if path is not None:
            re_path = re.compile(fnmatch.translate(path))
        else:
            re_path = None

        if sort is not None:
            if sort.startswith("-"):
                sort = sort[1:]
                sort_reverse = True
            else:
                sort_reverse = False

        pages = []
        for page in self.site.pages.values():
            if not page.FINDABLE: continue
            if re_path is not None and not re_path.match(page.src_relpath): continue
            if sort is not None and sort not in page.meta: continue
            pages.append(page)

        if sort is not None:
            pages.sort(key=lambda p: p.meta.get(sort, None), reverse=sort_reverse)

        if limit is not None:
            pages = pages[:limit]

        return pages




#!/usr/bin/env python3

from distutils.core import setup
import sys

setup(
    name = "staticsite",
    requires=[ 'unidecode', 'markdown', 'toml', 'yaml', 'jinja2', 'dateutil', 'livereload' ],
    version = "0.4",
    description = "Static site generator",
    author = ["Enrico Zini"],
    author_email = ["enrico@enricozini.org"],
    url = "https://github.com/spanezz/staticsite",
    license = "http://www.gnu.org/licenses/gpl-3.0.html",
    packages = ["staticsite"],
    scripts = ['ssite']
)



# coding: utf-8

from .core import Page, settings, RenderedString
import os
import re
from collections import defaultdict
import jinja2
import logging

log = logging.getLogger()

class DirPage(Page):
    """
    A directory index
    """
    TYPE = "dir"
    ANALYZE_PASS = 3
    RENDER_PREFERRED_ORDER = 2

    def __init__(self, site, relpath, pages):
        super().__init__(
            site=site,
            root_abspath=None,
            src_relpath=relpath,
            src_linkpath=relpath,
            dst_relpath=os.path.join(relpath, "index.html"),
            dst_link=os.path.join(settings.SITE_ROOT, relpath))

        self.pages = list(pages)
        self.subdirs = []

    def attach_to_parent(self):
        if not self.src_relpath: return
        parent_relpath = os.path.dirname(self.src_relpath)
        parent = self.site.pages[parent_relpath]
        if parent.TYPE != "dir": return
        if self in parent.subdirs: return
        parent.subdirs.append(self)
        parent.attach_to_parent()

    @property
    def src_abspath(self):
        return None

    def get_date(self):
        # Sort by decreasing date
        res = self.meta.get("date", None)
        if res is None:
            self.pages.sort(key=lambda x:x.meta["date"], reverse=True)
            if self.pages:
                dates = [self.pages[0].meta["date"]]
            else:
                dates = []
            dates.extend(d.get_date() for d in self.subdirs)
            self.meta["date"] = res = max(dates)
        return res

    def read_metadata(self):
        self.meta["date"] = self.get_date()
        self.meta["title"] = os.path.basename(self.src_relpath) or settings.SITE_NAME

    def render(self):
        self.subdirs.sort(key=lambda x:x.meta["title"])
        parent_page = None
        if self.src_relpath:
            parent = os.path.dirname(self.src_relpath)
            parent_page = self.site.pages.get(parent, None)

        body = self.site.theme.dir_template.render(
            parent_page=parent_page,
            page=self,
            pages=self.subdirs + self.pages,
        )
        return {
            self.dst_relpath: RenderedString(body)
        }



# coding: utf-8
import os
import logging

log = logging.getLogger()

class Archetypes:
    def __init__(self, site, root):
        self.site = site

        # Root directory where archetypes are found
        self.root = root

    def find(self, name):
        """
        Read the archetypes directory and return the archetype that matches the given name.

        Returns None if nothing matches.
        """
        # Map input file patterns to resource handlers
        from .markdown import MarkdownPages
        handlers = [
            # Only resources that have a front matter can be used here, because
            # we need to read the front matter to determine the destination
            # path
            MarkdownPages(self.site),
        ]

        for root, dnames, fnames in os.walk(self.root):
            for f in fnames:
                if f.startswith("."): continue
                relpath = os.path.relpath(os.path.join(root, f), self.root)
                for handler in handlers:
                    a = handler.try_load_archetype(self, relpath, name)
                    if a is not None:
                        return a
        return None



# coding: utf-8
import os
import sys
import logging
import shutil
import mimetypes
from . import content

log = logging.getLogger()

class Settings:
    def __init__(self):
        from . import global_settings
        self.add_module(global_settings)

    def as_dict(self):
        res = {}
        for setting in dir(self):
            if setting.isupper():
                res[setting] = getattr(self, setting)
        return res

    def add_module(self, mod):
        """
        Add uppercase settings from mod into this module
        """
        for setting in dir(mod):
            if setting.isupper():
                setattr(self, setting, getattr(mod, setting))

    def load(self, pathname):
        """
        Load settings from a python file, importing only uppercase symbols
        """
        orig_dwb = sys.dont_write_bytecode
        try:
            sys.dont_write_bytecode = True
            # http://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path

            # Seriously, this should not happen in the standard library. You do not
            # break stable APIs. You can extend them but not break them. And
            # especially, you do not break stable APIs and then complain that people
            # stick to 2.7 until its death, and probably after.
            if sys.version_info >= (3, 5):
                import importlib.util
                spec = importlib.util.spec_from_file_location("staticsite.settings", pathname)
                user_settings = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(user_settings)
            else:
                from importlib.machinery import SourceFileLoader
                user_settings = SourceFileLoader("staticsite.settings", pathname).load_module()
        finally:
            sys.dont_write_bytecode = orig_dwb

        self.add_module(user_settings)


settings = Settings()



class Archetype:
    def __init__(self, site, relpath):
        self.site = site
        self.relpath = relpath

    def as_template(self, **kw):
        abspath = os.path.join(site.archetypes_root, self.relpath)
        with open(abspath, "rt") as fd:
            return self.site.jinja2.from_string(fd.read(), **kw)


class Page:
    """
    A source page in the site.

    This can be a static asset, a file to be rendered, a taxonomy, a
    directory listing, or anything else.

    All pages have a number of members to describe their behaviour in the site:

    `site`:
        the Site that contains the page
    `root_abspath`:
        the absolute path of the root of the directory tree under which the
        page was found. This is where src_relpath is rooted.
    `src_relpath`:
        the relative path under `root_abspath` for the source file of the page
    `src_linkpath`:
        the path used by other pages to link to this page, when referencing its
        source. For example, blog/2016/example.md is linked as
        blog/2016/example.
    `dst_relpath`:
        relative path of the file that will be generated for this page when it
        gets rendered.
        FIXME: now that a resource can generate multiple files, this is not
        really needed.
    `dst_link`:
        absolute path in the site namespace used to link to this page in
        webpages.
        FIXME: now that a resource can generate multiple files, this is not
        really needed.
    `meta`:
        a dictionary with the page metadata. See the README for documentation
        about its contents.
    """
    # In what pass must pages of this type be analyzed.
    ANALYZE_PASS = 1

    # True if the page can be found when search site contents
    FINDABLE = False

    # Preferred order of rendering, not for functional purposes for for
    # purposes of collecting reasonable timings. This can be used for example
    # to render markdown pages before taxonomies, so that the time of rendering
    # markdown is counted as such and not as time needed for rendering
    # taxonomies.
    RENDER_PREFERRED_ORDER = 1

    def __init__(self, site, root_abspath, src_relpath, src_linkpath, dst_relpath, dst_link):
        self.site = site
        self.root_abspath = root_abspath
        self.src_relpath = src_relpath
        self.src_linkpath = src_linkpath
        self.dst_relpath = dst_relpath
        self.dst_link = dst_link
        self.meta = {}
        log.debug("%s: new page, src_link: %s", src_relpath, src_linkpath)

    @property
    def src_abspath(self):
        return os.path.join(self.root_abspath, self.src_relpath)

    @property
    def date_as_iso8601(self):
        from dateutil.tz import tzlocal
        ts = self.meta.get("date", None)
        if ts is None: return None
        # TODO: Take timezone from config instead of tzlocal()
        tz = tzlocal()
        ts = ts.astimezone(tz)
        offset = tz.utcoffset(ts)
        offset_sec = (offset.days * 24 * 3600 + offset.seconds)
        offset_hrs = offset_sec // 3600
        offset_min = offset_sec % 3600
        if offset:
            tz_str = '{0:+03d}:{1:02d}'.format(offset_hrs, offset_min // 60)
        else:
            tz_str = 'Z'
        return ts.strftime("%Y-%m-%d %H:%M:%S") + tz_str

    def resolve_link(self, target):
        dirname, basename = os.path.split(target)
        if basename == "index.html":
            #log.debug("%s: resolve %s using %s", self.src_relpath, target, dirname)
            target = dirname
        #else:
            #log.debug("%s: resolve %s using %s", self.src_relpath, target, target)

        # Absolute URLs are resolved as is
        if target.startswith("/"):
            if target == "/":
                target_relpath = ""
            else:
                target_relpath = os.path.normpath(target.lstrip("/"))
            #log.debug("%s: resolve absolute path using %s", self.src_relpath, target_relpath)
            return self.site.pages.get(target_relpath, None)

        root = os.path.dirname(self.src_relpath)
        while True:
            target_relpath = os.path.normpath(os.path.join(root, target))
            res = self.site.pages.get(target_relpath, None)
            if res is not None: return res
            if not root or root == "/":
                return None
            root = os.path.dirname(root)

    def read_metadata(self):
        """
        Fill in self.meta scanning the page contents
        """
        pass

    def check(self, checker):
        pass

    def target_relpaths(self):
        return [self.dst_relpath]

    def __str__(self):
        return self.src_relpath

    def __repr__(self):
        return "{}:{}".format(self.TYPE, self.src_relpath)


class RenderedFile:
    def __init__(self, abspath):
        self.abspath = abspath

    def write(self, dst):
        shutil.copy2(self.abspath, dst)

    def content(self):
        with open(self.abspath, "rb") as fd:
            return fd.read()


class RenderedString:
    def __init__(self, s):
        self.buf = s.encode("utf-8")

    def write(self, dst):
        with open(dst, "wb") as out:
            out.write(self.buf)

    def content(self):
        return self.buf


class PageFS:
    """
    VFS-like abstraction that maps the names of files that pages would render
    with the corresponding pages.

    This can be used to render pages on demand.
    """
    def __init__(self):
        self.paths = {}

    def add_site(self, site):
        for page in site.pages.values():
            for relpath in page.target_relpaths():
                self.add_page(page, relpath)

    def add_page(self, page, dst_relpath=None):
        if dst_relpath is None:
            dst_relpath = page.dst_relpath
        self.paths[dst_relpath] = page

    def get_page(self, relpath):
        if not relpath:
            relpath = "/index.html"
        dst_relpath = os.path.normpath(relpath).lstrip("/")
        page = self.paths.get(dst_relpath, None)
        if page is not None:
            return dst_relpath, page

        dst_relpath = os.path.join(dst_relpath, "index.html")
        page = self.paths.get(dst_relpath, None)
        if page is not None:
            return dst_relpath, page

        return None, None

    def serve_path(self, path, environ, start_response):
        """
        Render a page on the fly and serve it.

        Call start_response with the page headers and return the bytes() with
        the page contents.

        start_response is the start_response from WSGI

        Returns None without calling start_response if no page was found.
        """
        dst_relpath, page = self.get_page(os.path.normpath(path).lstrip("/"))
        if page is None: return None

        for relpath, rendered in page.render().items():
            if relpath == dst_relpath:
                break
        else:
            return None

        content = rendered.content()
        start_response("200 OK", [
            ("Content-Type", mimetypes.guess_type(relpath)[0]),
            ("Content-Length", str(len(content))),
        ])
        return [content]
